//! Implementation of [`MonadFail<E>`] for [`Result<A, E>`].
//!
//! If any of the input values are [`Err`], you can expect the output to be [`Err`] as well.
//! That includes
//! [`ResultClass::replace`] and [`ResultClass::discard_first`]/[`ResultClass::discard_second`],
//! even if the value of the option would be ignored.
//!
//! For [`Option<A>`] alternative see [`super::option`]

use crate::func::*;

pub struct ResultClass<E: 'static>(E);

impl<E> WeakFunctor for ResultClass<E> {
    type F<'a, A: 'a> = Result<A, E>;
}

impl<E> Functor for ResultClass<E> {
    fn fmap<'a, A: 'a, B: 'a>(f: impl 'a + FnOnce(A) -> B, fa: Self::F<'a, A>) -> Self::F<'a, B> {
        fa.map(f)
    }

    fn replace<'a, A: 'a, B: 'a>(fa: Self::F<'a, A>, b: B) -> Self::F<'a, B> {
        fa.map(|_| b)
    }
}

impl<E> Pure for ResultClass<E> {
    fn pure<'a, A: 'a>(a: A) -> Self::F<'a, A> {
        Ok(a)
    }
}

impl<E> ApplicativeSeq for ResultClass<E> {
    fn seq<'a, A: 'a, B: 'a>(
        ff: Self::F<'a, impl 'a + FnOnce(A) -> B>,
        fa: Self::F<'a, A>,
    ) -> Self::F<'a, B> {
        Self::pure(ff?(fa?))
    }
}

impl<E> ApplicativeLA2 for ResultClass<E> {
    fn la2<'a, A: 'a, B: 'a, C: 'a>(
        f: impl 'a + FnOnce(A, B) -> C,
        fa: Self::F<'a, A>,
        fb: Self::F<'a, B>,
    ) -> Self::F<'a, C> {
        Self::pure(f(fa?, fb?))
    }
}

impl<E> ApplicativeTuple for ResultClass<E> {
    fn tuple<'a, A: 'a, B: 'a>((fa, fb): (Self::F<'a, A>, Self::F<'a, B>)) -> Self::F<'a, (A, B)>
    where
        Self: 'a,
    {
        Self::pure((fa?, fb?))
    }
}

impl<E> Applicative for ResultClass<E> {
    fn discard_first<'a, A: 'a, B: 'a>(fa: Self::F<'a, A>, fb: Self::F<'a, B>) -> Self::F<'a, B> {
        fa?;
        fb
    }

    fn discard_second<'a, A: 'a, B: 'a>(fa: Self::F<'a, A>, fb: Self::F<'a, B>) -> Self::F<'a, A> {
        fb?;
        fa
    }
}

impl<E> Monad for ResultClass<E> {
    fn bind<'a, A: 'a, B: 'a>(
        fa: Self::F<'a, A>,
        f: impl 'a + FnOnce(A) -> Self::F<'a, B>,
    ) -> Self::F<'a, B> {
        f(fa?)
    }

    fn ibind<'a, A: 'a, B: 'a>(
        mut a: A,
        mut f: impl 'a + FnMut(A) -> Self::F<'a, IState<A, B>>,
    ) -> Self::F<'a, B>
    where
        Self: 'a,
    {
        loop {
            match f(a)? {
                IState::Pending(next_a) => a = next_a,
                IState::Done(b) => return Self::pure(b),
            };
        }
    }

    fn join<'a, A: 'a>(ffa: Self::F<'a, Self::F<'a, A>>) -> Self::F<'a, A> {
        ffa?
    }
}

impl<E> MonadFail<E> for ResultClass<E> {
    fn fail<'a, A: 'a>(e: E) -> Self::F<'a, A>
    where
        Self: 'a,
    {
        Err(e)
    }
}
