use crate::std::inlining::*;

use super::*;

impl Serializable for bool {
    fn serialize(&self, serializer: &mut dyn Serializer) {
        match self {
            true => serializer.write(&[0]),
            false => serializer.write(&[1]),
        }
    }
}

#[derive(Debug)]
pub enum BooleanParseError {
    OutOfBounds(u8),
    Eof,
    ExtraData(usize),
}

impl Display for BooleanParseError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            BooleanParseError::OutOfBounds(value) => {
                f.write_fmt(format_args!("boolean value out of bounds: {}.", value))
            }
            BooleanParseError::Eof => {
                f.write_fmt(format_args!("encountered EOF write parsing a boolean."))
            }
            BooleanParseError::ExtraData(length) => f.write_fmt(format_args!(
                "encountered extra data of length {} while parsing a boolean",
                length
            )),
        }
    }
}

impl Error for BooleanParseError {}

impl From<&[u8]> for BooleanParseError {
    fn from(_value: &[u8]) -> Self {
        Self::Eof
    }
}

impl Atomic for bool {
    type ParseError = BooleanParseError;

    fn deserialize(deserializer: &mut dyn Deserializer) -> Result<Self, Self::ParseError> {
        let byte = deserializer.read_n_const::<1>()?;
        match byte[0] {
            0 => Ok(false),
            1 => Ok(true),
            value => Err(BooleanParseError::OutOfBounds(value)),
        }
    }

    fn unexpected_tail(tail: &[u8]) -> Self::ParseError {
        BooleanParseError::ExtraData(tail.len())
    }
}

impl ConstSizeAtomic for bool {
    const SIZE: usize = 1;
}
